""" Event detection CRF finetuning: utilities to work with ACE 2005 & DuEE """

from __future__ import absolute_import, division, print_function
import json
import logging
import os
import sys
from copy import deepcopy
from io import open
from transformers import XLMRobertaTokenizer, BertTokenizer, RobertaTokenizer, AutoTokenizer
from .utils_ee import read_by_lines, write_by_lines, load_dict, label_data, process_remained_pred_trigger, clear_wrong_tokens

from torch.nn.utils.rnn import pad_sequence
from tqdm import tqdm

logger = logging.getLogger(__name__)

class InputExample(object):
    """A single training/test example for token classification."""

    def __init__(self, guid, words, labels, trigger_word_ids):
        """Constructs a InputExample.

        Args:
            guid: Unique id for the example.
            words: list. The words of the sequence.
            labels: (Optional) list. The labels for each word of the sequence. This should be
            specified for train and dev examples, but not for test examples.
            trigger_word_ids: set token_type_ids of trigger tokens to 1, used in argument extraction task
        """
        self.guid = guid
        self.words = words
        self.labels = labels
        self.trigger_word_ids = trigger_word_ids


class InputFeatures(object):
    """A single set of features of data."""

    def __init__(self, input_ids, input_mask, segment_ids, label_ids, token_type_ids):
        self.input_ids = input_ids
        self.input_mask = input_mask
        self.segment_ids = segment_ids
        self.label_ids = label_ids
        self.token_type_ids = token_type_ids

class DataProcessor(object):

    def __init__(self, task_name, tokenizer):
        self.task_name = task_name
        self.tokenizer = tokenizer

    """Base class for data converters for multiple choice data sets."""

    def get_examples(self, file_path, set_type):
        """Gets a collection of `InputExample`s for the train/dev/test set."""
        raise NotImplementedError()

    def get_labels(self):
        """Gets the list of labels for this data set."""
        raise NotImplementedError()
    
    def process_dev_with_pred_trigger(self):
        """Process the dev set with pred trigger."""
        raise NotImplementedError()
    
    def process_test_with_pred_trigger(self):
        """Process the dev set with test trigger."""
        raise NotImplementedError()
    
    
class DUEEProcessor(DataProcessor):
    """Processor for the DuEE data set."""
    
    def get_examples(self, file_path, set_type):
        """See base class."""
        logger.info("LOOKING AT {} train".format(file_path))
        return self._create_examples(file_path, set_type)
    
    def data_process(self, path, trigger_pred_list, model="role"):

        output = ["text_a\tlabel\ttrigger_tag\tindex"]
        with open(path) as f:
            raw = f.readlines()
            assert len(raw) == len(trigger_pred_list)
            for idx, line, trigger_pred,in tqdm(zip(range(len(raw)), raw, trigger_pred_list)):

                d_json = json.loads(line.strip())
                text_a = ["，" if t == " " or t == "\n" or t == "\t" else t for t in list(d_json["text"].lower())]
                text_a_cleaned = text_a

                if model == "role":
                    for event in d_json.get("event_list", []):
                        event_type = event["event_type"]
                        trigger_labels = ["O"] * len(text_a)
                        trigger_start = event["trigger_start_index"]
                        trigger = event["trigger"]
                        trigger_labels = label_data(trigger_labels, trigger_start, len(trigger), event_type)

                        text_a_cleaned, trigger_labels_cleaned, trigger_start = clear_wrong_tokens(text_a, trigger_labels, self.tokenizer, len(trigger_pred))

                        for i in range(trigger_start, trigger_start + len(trigger)):
                            if i == len(trigger_pred): break
                            trigger_labels_cleaned[i] = trigger_pred[i]
                            trigger_pred[i] = "O"

                        role_labels = ["O"] * len(text_a)
                        for arg in event["arguments"]:
                            role_type = arg["role"]
                            argument = arg["argument"]
                            role_start = arg["argument_start_index"]
                            role_labels = label_data(role_labels, role_start, len(argument), role_type)

                        text_a_cleaned, role_labels_cleaned, _ = clear_wrong_tokens(text_a, role_labels, self.tokenizer, len(trigger_pred))

                        assert len(text_a_cleaned) == len(role_labels_cleaned) == len(trigger_labels_cleaned)
                        output.append("{}\t{}\t{}\t{}".format("\002".join(text_a_cleaned), "\002".join(role_labels_cleaned), "\002".join(trigger_labels_cleaned), idx))

                    # print("trigger_pred_after:", trigger_pred)
                    trigger_remained_labels_list = process_remained_pred_trigger(trigger_pred)
                    
                    for remained_labels in trigger_remained_labels_list:
                        assert len(text_a_cleaned) == len(remained_labels)
                        output.append("{}\t{}\t{}\t{}".format("\002".join(text_a_cleaned), "\002".join(["O"] * len(text_a_cleaned)), "\002".join(remained_labels), idx))
                
        return output    

    
    def process_dev_with_pred_trigger(self, args, raw_path, out_file):
        raw_path = raw_path + "/raw"
        trigger_pred_list = json.load(open(args.dev_trigger_pred_file, "r"))
        dev_role = self.data_process(os.path.join(raw_path, "duee_dev.json"), trigger_pred_list, "role")
        write_by_lines(os.path.join(args.data_dir, out_file), dev_role)
    
    def process_test_with_pred_trigger(self, args, raw_path, out_file):
        '''Do not consider the test set in duee.'''
        pass

    def get_labels(self, tag_path):
        """See base class."""
        tag_path = os.path.join(tag_path, self.task_name + "_tag.dict")
        return load_dict(tag_path).keys()
    
    def _create_examples(self, file_path, set_type):
        """Creates examples for the training and dev sets."""

        examples = []   

        for (idx, line) in enumerate(open(file_path).readlines()[1:]):

            e_id = "%s-%s" % (set_type, idx)

            datas = line.strip('\n').split('\t')
            words = datas[0].split('\002')
            labels = datas[1].split('\002')
            trigger_word_ids = None if self.task_name == "trigger" else datas[2].split('\002')

            examples.append(
                InputExample(
                    guid=e_id,
                    words=words,
                    labels=labels,
                    trigger_word_ids=trigger_word_ids
                )
            )

        return examples



class ACEProcessor(DataProcessor):

    """Processor for the DuEE data set."""
    
    def get_examples(self, file_path, set_type):
        """See base class."""
        logger.info("LOOKING AT {} train".format(file_path))
        return self._create_examples(file_path, set_type)
    

    def data_process(self, raw_data, trigger_pred_list):
        
        output = ["text_a\tlabel\ttrigger_tag\tindex"]

        assert len(raw_data) == len(trigger_pred_list)
        for idx, line in enumerate(tqdm(raw_data, desc="processing pred_trigger")):

            line = json.loads(line)
            text_a = line["tokens"]
            trigger_pred = trigger_pred_list[idx]

            assert len(trigger_pred) == len(text_a)

            if len(line["event_mentions"]) == 0:
                output.append("{}\t{}\t{}\t{}".format(" ".join(text_a), " ".join(["O"] * len(text_a)), " ".join(trigger_pred), idx))
                continue

            entity_mention = line["entity_mentions"]
            entity_id2mention = {mention["id"]: mention for mention in entity_mention}

            for mention in line["event_mentions"]:

                trigger_labels = ["O"] * len(text_a)
                event_type = mention["event_type"]
                trigger_start = mention["trigger"]["start"]
                trigger_end = mention["trigger"]["end"]
                trigger = mention["trigger"]["text"]
                trigger_labels = label_data(trigger_labels, trigger_start, trigger_end - trigger_start, event_type)

                for i in range(trigger_start, trigger_end):
                    if i == len(trigger_pred): break
                    trigger_labels[i] = trigger_pred[i]
                    trigger_pred[i] = "O"

                role_labels = ["O"] * len(text_a)
                for arg in mention["arguments"]:
                    role_type = arg["role"]
                    argument = arg["text"]
                    role_entity = entity_id2mention[arg["entity_id"]]
                    role_start = role_entity["start"]
                    role_end = role_entity["end"]
                    role_labels = label_data(role_labels, role_start, role_end - role_start, role_type)

                assert len(text_a) == len(role_labels) == len(trigger_labels)
                output.append("{}\t{}\t{}\t{}".format(" ".join(text_a), " ".join(role_labels), " ".join(trigger_labels), idx))

            trigger_remained_labels_list = process_remained_pred_trigger(trigger_pred)
            for remained_labels in trigger_remained_labels_list:
                assert len(text_a) == len(remained_labels)
                output.append("{}\t{}\t{}\t{}".format(" ".join(text_a), " ".join(["O"] * len(text_a)), " ".join(remained_labels), idx))

        print(len(output))
        return output    

    def process_dev_with_pred_trigger(self, args, raw_path, out_file):
        trigger_pred_list = json.load(open(args.dev_trigger_pred_file, "r"))
        raw_path = raw_path + "/degree"
        raw_data = open(os.path.join(raw_path, "dev.w1.oneie.json"), "r").readlines()
        dev_role = self.data_process(raw_data, trigger_pred_list)
        write_by_lines(os.path.join(args.data_dir, out_file), dev_role)
        
    
    def process_test_with_pred_trigger(self, args, raw_path, out_file):
        trigger_pred_list = json.load(open(args.test_trigger_pred_file, "r"))
        raw_path = raw_path + "/degree"
        raw_data = open(os.path.join(raw_path, "test.w1.oneie.json"), "r").readlines()
        test_role = self.data_process(raw_data, trigger_pred_list)
        write_by_lines(os.path.join(args.data_dir, out_file), test_role)

    def get_labels(self, tag_path):
        """See base class."""
        tag_path = os.path.join(tag_path, self.task_name + "_tag.json")
        return json.load(open(tag_path, "r")).keys()
    
    def _create_examples(self, file_path, set_type):
        """Creates examples for the training and dev sets."""

        examples = []   

        for (idx, line) in enumerate(open(file_path).readlines()[1:]):

            e_id = "%s-%s" % (set_type, idx)

            datas = line.strip('\n').split('\t')
            words = datas[0].split()
            labels = datas[1].split()
            trigger_word_ids = None if self.task_name == "trigger" else datas[2].split()

            examples.append(
                InputExample(
                    guid=e_id,
                    words=words,
                    labels=labels,
                    trigger_word_ids=trigger_word_ids
                )
            )
        return examples



def convert_examples_to_features(examples,
                                 label_list,
                                 max_seq_length,
                                 tokenizer,
                                 cls_token_at_end=False,
                                 cls_token="[CLS]",
                                 cls_token_segment_id=1,
                                 sep_token="[SEP]",
                                 sep_token_extra=False,
                                 pad_on_left=False,
                                 pad_token=0,
                                 pad_token_segment_id=0,
                                 pad_token_label_id=-100,
                                 sequence_a_segment_id=0,
                                 mask_padding_with_zero=True,
                                 model_name=None):
    """ Loads a data file into a list of `InputBatch`s
        `cls_token_at_end` define the location of the CLS token:
            - False (Default, BERT/XLM pattern): [CLS] + A + [SEP] + B + [SEP]
            - True (XLNet/GPT pattern): A + [SEP] + B + [SEP] + [CLS]
        `cls_token_segment_id` define the segment id associated to the CLS token (0 for BERT, 2 for XLNet)
    """

    label_map = {label: i for i, label in enumerate(label_list)}

    # my logic in crf_padding requires this check. I create mask for crf by labels==pad_token_label_id to not include it
    # in loss and decoding
    assert pad_token_label_id not in label_map.values()

    features = []
    for (ex_index, example) in enumerate(examples):
        if ex_index % 10000 == 0:
            print("###############")
            logger.info("Writing example %d of %d", ex_index, len(examples))
            print("###############")

        tokens = []
        label_ids = []
        token_type_ids = []

        assert len(example.words) == len(example.labels)
        if example.trigger_word_ids is not None:
            assert len(example.words) == len(example.labels) == len(example.trigger_word_ids)

        # print("example.words:", example.words)
        # print("example.labels:", example.labels)
        # print("example.trigger_word_ids:", example.trigger_word_ids)

        for idx, word, label in zip(range(len(example.words)), example.words, example.labels):
            word_tokens = tokenizer.tokenize(word)
            # clear out wrong chars
            if len(word_tokens) == 0: continue
            tokens.append(word)

            # Use the real label id for the first token of the word, and padding ids for the remaining tokens
            if label!='X':
                label_ids.extend([label_map[label]] + [pad_token_label_id] * (len([word]) - 1))
            else:
                label_ids.extend([pad_token_label_id] + [pad_token_label_id] * (len([word]) - 1))

            if example.trigger_word_ids is not None:
                trigger_word_id = example.trigger_word_ids[idx]
                token_type = [0] if trigger_word_id == "O" else [1]
                token_type_ids.extend(token_type)
            else:
                token_type_ids.extend([0])
        
        # print("len(tokens):", len(tokens))
        # print("len(label_ids):", len(label_ids))
        # print("len(token_type_ids):", len(token_type_ids))

        # Account for [CLS] and [SEP] with "- 2" and with "- 3" for RoBERTa.
        special_tokens_count = 3 if sep_token_extra else 2
        if len(tokens) > max_seq_length - special_tokens_count:
            tokens = tokens[:(max_seq_length - special_tokens_count)]
            label_ids = label_ids[:(max_seq_length - special_tokens_count)]
            token_type_ids = token_type_ids[:(max_seq_length - special_tokens_count)]

        # The convention in BERT is:
        # (a) For sequence pairs:
        #  tokens:   [CLS] is this jack ##son ##ville ? [SEP] no it is not . [SEP]
        #  type_ids:   0   0  0    0    0     0       0   0   1  1  1  1   1   1
        # (b) For single sequences:
        #  tokens:   [CLS] the dog is hairy . [SEP]
        #  type_ids:   0   0   0   0  0     0   0
        #
        # Where "type_ids" are used to indicate whether this is the first
        # sequence or the second sequence. The embedding vectors for `type=0` and
        # `type=1` were learned during pre-training and are added to the wordpiece
        # embedding vector (and position vector). This is not *strictly* necessary
        # since the [SEP] token unambiguously separates the sequences, but it makes
        # it easier for the model to learn the concept of sequences.
        #
        # For classification tasks, the first vector (corresponding to [CLS]) is
        # used as as the "sentence vector". Note that this only makes sense because
        # the entire model is fine-tuned.
        tokens += [sep_token]
        label_ids += [pad_token_label_id]  # [label_map["X"]]
        token_type_ids += [0]
        if sep_token_extra:
            # roberta uses an extra separator b/w pairs of sentences
            tokens += [sep_token]
            label_ids += [pad_token_label_id]
            token_type_ids += [0]
        segment_ids = [sequence_a_segment_id] * len(tokens)

        if cls_token_at_end:
            tokens += [cls_token]
            label_ids += [pad_token_label_id]
            token_type_ids += [0]
            segment_ids += [cls_token_segment_id]
        else:
            tokens = [cls_token] + tokens
            token_type_ids += [0]
            label_ids = [pad_token_label_id] + label_ids
            segment_ids = [cls_token_segment_id] + segment_ids

        if model_name:
            if model_name == 'xlm-roberta-base':
                tokenizer = XLMRobertaTokenizer.from_pretrained(model_name)
                input_ids = tokenizer.convert_tokens_to_ids(tokens)
            elif model_name.startswith('bert'):
                tokenizer = BertTokenizer.from_pretrained(model_name)
                input_ids = tokenizer.convert_tokens_to_ids(tokens)
            elif model_name == 'roberta':
                tokenizer = RobertaTokenizer.from_pretrained(model_name)
                input_ids = tokenizer.convert_tokens_to_ids(tokens)
        else:
            input_ids = tokenizer.convert_tokens_to_ids(tokens)

        # The mask has 1 for real tokens and 0 for padding tokens. Only real
        # tokens are attended to.
        input_mask = [1 if mask_padding_with_zero else 0] * len(input_ids)

        # Zero-pad up to the sequence length.
        padding_length = max_seq_length - len(input_ids)
        if pad_on_left:
            input_ids = ([pad_token] * padding_length) + input_ids
            input_mask = ([0 if mask_padding_with_zero else 1] * padding_length) + input_mask
            segment_ids = ([pad_token_segment_id] * padding_length) + segment_ids
            label_ids = ([pad_token_label_id] * padding_length) + label_ids
            token_type_ids = ([0] * padding_length) + token_type_ids
        else:
            input_ids += ([pad_token] * padding_length)
            input_mask += ([0 if mask_padding_with_zero else 1] * padding_length)
            segment_ids += ([pad_token_segment_id] * padding_length)
            label_ids += ([pad_token_label_id] * padding_length)
            token_type_ids += ([0] * padding_length)

        # print("len(input_ids):", len(input_ids))
        # print("len(label_ids):", len(label_ids))
        # print("len(token_type_ids):", len(token_type_ids))
        assert len(input_ids) == max_seq_length
        assert len(input_mask) == max_seq_length
        assert len(segment_ids) == max_seq_length
        assert len(label_ids) == max_seq_length
        assert len(token_type_ids) == max_seq_length

        if ex_index < 0:
            logger.info("*** Example ***")
            logger.info("guid: %s", example.guid)
            logger.info("tokens: %s", " ".join([str(x) for x in tokens]))
            logger.info("input_ids: %s", " ".join([str(x) for x in input_ids]))
            logger.info("input_mask: %s", " ".join([str(x) for x in input_mask]))
            logger.info("segment_ids: %s", " ".join([str(x) for x in segment_ids]))
            logger.info("label_ids: %s", " ".join([str(x) for x in label_ids]))
            logger.info("token_type_ids: %s", " ".join([str(x) for x in token_type_ids]))

        features.append(
            InputFeatures(input_ids=input_ids,
                          input_mask=input_mask,
                          segment_ids=segment_ids,
                          label_ids=label_ids,
                          token_type_ids=token_type_ids))
    return features


PROCESSORS = {
    "DuEE":DUEEProcessor, 
    "ACE":ACEProcessor
              }


if __name__ == "__main__":
    pad_token_label_id = -100
    tokenizer = AutoTokenizer.from_pretrained("../../models/bert-base-uncased")
    processor = PROCESSORS["ACE"]("role", tokenizer)
    # labels = processor.get_labels("./data/DuEE/schema/")
    processor.process_test_with_pred_trigger("./data/ACE/bertcrf/role", "./exp/ACE/trigger_bert-base-uncased/test_pred.json")


    # features = convert_examples_to_features(
    #         train_examples, 
    #         labels, 
    #         128, tokenizer,
    #         cls_token_at_end=bool("bertcrf" in ["xlnet"]),# xlnet has a cls token at the end
    #         cls_token=tokenizer.cls_token,
    #         cls_token_segment_id=2 if "bertcrf" in ["xlnet"] else 0,
    #         sep_token=tokenizer.sep_token,
    #         sep_token_extra=bool("bertcrf" in ["roberta"]),# roberta uses an extra separator b/w pairs of sentences, cf. github.com/pytorch/fairseq/commit/1684e166e3da03f5b600dbb7855cb98ddfcd0805
    #         pad_on_left=bool("bertcrf" in ["xlnet"]), # pad on the left for xlnet
    #         pad_token=tokenizer.convert_tokens_to_ids([tokenizer.pad_token])[0],
    #         pad_token_segment_id=4 if "bertcrf" in ["xlnet"] else 0,
    #         pad_token_label_id=pad_token_label_id
    #     )




from typing import Any, Optional, Tuple
import math

import copy
import torch
from torch import nn, Tensor, device

from transformers.activations import ACT2FN
from transformers.modeling_utils import (
    PreTrainedModel,
    apply_chunking_to_forward,
)
from transformers.configuration_utils import PretrainedConfig
from transformers.modeling_outputs import BaseModelOutput, BaseModelOutputWithPooling

# some function
def get_extended_attention_mask(attention_mask: Tensor, input_shape: Tuple[int], device: device) -> Tensor:
        """
        Makes broadcastable attention and causal masks so that future and masked tokens are ignored.

        Arguments:
            attention_mask (:obj:`torch.Tensor`):
                Mask with ones indicating tokens to attend to, zeros for tokens to ignore.
            input_shape (:obj:`Tuple[int]`):
                The shape of the input to the model.
            device: (:obj:`torch.device`):
                The device of the input to the model.

        Returns:
            :obj:`torch.Tensor` The extended attention mask, with a the same dtype as :obj:`attention_mask.dtype`.
        """
        # We can provide a self-attention mask of dimensions [batch_size, from_seq_length, to_seq_length]
        # ourselves in which case we just need to make it broadcastable to all heads.
        if attention_mask.dim() == 3:
            extended_attention_mask = attention_mask[:, None, :, :]
        elif attention_mask.dim() == 2:
            # Provided a padding mask of dimensions [batch_size, seq_length]
            # - if the model is a decoder, apply a causal mask in addition to the padding mask
            # - if the model is an encoder, make the mask broadcastable to [batch_size, num_heads, seq_length, seq_length]
            extended_attention_mask = attention_mask[:, None, None, :]
        else:
            raise ValueError(
                f"Wrong shape for input_ids (shape {input_shape}) or attention_mask (shape {attention_mask.shape})"
            )

        # Since attention_mask is 1.0 for positions we want to attend and 0.0 for
        # masked positions, this operation will create a tensor which is 0.0 for
        # positions we want to attend and -10000.0 for masked positions.
        # Since we are adding it to the raw scores before the softmax, this is
        # effectively the same as removing these entirely.
        extended_attention_mask = extended_attention_mask.to(dtype=torch.long)  # fp16 compatibility
        extended_attention_mask = (1.0 - extended_attention_mask) * -10000.0
        return extended_attention_mask


def get_head_mask(
        head_mask: Optional[Tensor], num_hidden_layers: int, is_attention_chunked: bool = False
    ) -> Tensor:
        """
        Prepare the head mask if needed.

        Args:
            head_mask (:obj:`torch.Tensor` with shape :obj:`[num_heads]` or :obj:`[num_hidden_layers x num_heads]`, `optional`):
                The mask indicating if we should keep the heads or not (1.0 for keep, 0.0 for discard).
            num_hidden_layers (:obj:`int`):
                The number of hidden layers in the model.
            is_attention_chunked: (:obj:`bool`, `optional`, defaults to :obj:`False`):
                Whether or not the attentions scores are computed by chunks or not.

        Returns:
            :obj:`torch.Tensor` with shape :obj:`[num_hidden_layers x batch x num_heads x seq_length x seq_length]` or
            list with :obj:`[None]` for each layer.
        """
        head_mask = [None] * num_hidden_layers

        return head_mask


# models
class IFAConfig(PretrainedConfig):
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class IFAPreTrainedModel(PreTrainedModel):
    config_class = IFAConfig
    base_model_prefix = "clip"
    supports_gradient_checkpointing = True
    _keys_to_ignore_on_load_missing = [r"position_ids"]
    
    def __init_weights(self, module):
        pass


class CLIPVisionEmbeddings(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.config = config
        self.embed_dim = config.hidden_size
        self.image_size = config.image_size
        self.patch_size = config.patch_size

        self.class_embedding = nn.Parameter(torch.randn(self.embed_dim))

        self.patch_embedding = nn.Conv2d(
            in_channels=3, out_channels=self.embed_dim, kernel_size=self.patch_size, stride=self.patch_size, bias=False
        )

        self.num_patches = (self.image_size // self.patch_size) ** 2
        self.num_positions = self.num_patches + 1
        self.position_embedding = nn.Embedding(self.num_positions, self.embed_dim)
        self.register_buffer("position_ids", torch.arange(self.num_positions).expand((1, -1)))

        self.aux_position_embedding = nn.Embedding(48, self.embed_dim)
        self.register_buffer("aux_position_ids", torch.arange(48).expand((1, -1)))

        self.rcnn_position_embedding = nn.Embedding(12, self.embed_dim)
        self.register_buffer("rcnn_position_ids", torch.arange(12).expand((1, -1)))

    def forward(self, pixel_values, aux_embeddings=None, rcnn_embeddings=None):
        batch_size = pixel_values.shape[0]

        class_embeds = self.class_embedding.expand(batch_size, 1, -1)
        embeddings = class_embeds

        if aux_embeddings is not None:
            aux_embeds = []
            for aux_embedding in aux_embeddings:
                aux_embed = self.patch_embedding(aux_embedding)
                aux_embed = aux_embed.flatten(2).transpose(1, 2).flatten(0, 1)    # 3*16, 768 3个子图
                aux_embeds.append(aux_embed)
            aux_embeds = torch.stack(aux_embeds) # bsz, 48, 768
            aux_embeds = aux_embeds + self.aux_position_embedding(self.aux_position_ids)
            embeddings = torch.cat((embeddings, aux_embeds), dim=1)

        if rcnn_embeddings is not None:
            rcnn_embeds = []
            for rcnn_embedding in rcnn_embeddings:
                rcnn_embed = self.patch_embedding(rcnn_embedding)
                rcnn_embed = rcnn_embed.flatten(2).transpose(1, 2).flatten(0, 1)    # 3*4, 768 3个子图
                rcnn_embeds.append(rcnn_embed)
            rcnn_embeds = torch.stack(rcnn_embeds) # bsz, 12, 768
            rcnn_embeds = rcnn_embeds + self.rcnn_position_embedding(self.rcnn_position_ids)
            embeddings = torch.cat((embeddings, rcnn_embeds), dim=1)
        return embeddings


class BertEmbeddings(nn.Module):
    """Construct the embeddings from word, position and token_type embeddings."""

    def __init__(self, config):
        super().__init__()
        self.word_embeddings = nn.Embedding(config.vocab_size, config.hidden_size, padding_idx=config.pad_token_id)
        self.position_embeddings = nn.Embedding(config.max_position_embeddings, config.hidden_size)
        self.token_type_embeddings = nn.Embedding(config.type_vocab_size, config.hidden_size)

        # self.LayerNorm is not snake-cased to stick with TensorFlow model variable name and be able to load
        # any TensorFlow checkpoint file
        self.LayerNorm = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)
        # position_ids (1, len position emb) is contiguous in memory and exported when serialized
        self.position_embedding_type = getattr(config, "position_embedding_type", "absolute")
        self.register_buffer("position_ids", torch.arange(config.max_position_embeddings).expand((1, -1)))

    def forward(
        self, input_ids=None, token_type_ids=None, position_ids=None, inputs_embeds=None, past_key_values_length=0
    ):
        if input_ids is not None:
            input_shape = input_ids.size()
        else:
            input_shape = inputs_embeds.size()[:-1]

        seq_length = input_shape[1]

        if position_ids is None:
            position_ids = self.position_ids[:, past_key_values_length : seq_length + past_key_values_length]

        # Setting the token_type_ids to the registered buffer in constructor where it is all zeros, which usually occurs
        # when its auto-generated, registered buffer helps users when tracing the model without passing token_type_ids, solves
        # issue #5664
        if token_type_ids is None:
            if hasattr(self, "token_type_ids"):
                buffered_token_type_ids = self.token_type_ids[:, :seq_length]
                buffered_token_type_ids_expanded = buffered_token_type_ids.expand(input_shape[0], seq_length)
                token_type_ids = buffered_token_type_ids_expanded
            else:
                token_type_ids = torch.zeros(input_shape, dtype=torch.long, device=self.position_ids.device)

        if inputs_embeds is None:
            inputs_embeds = self.word_embeddings(input_ids)
        token_type_embeddings = self.token_type_embeddings(token_type_ids)

        embeddings = inputs_embeds + token_type_embeddings
        if self.position_embedding_type == "absolute":
            position_embeddings = self.position_embeddings(position_ids)
            embeddings += position_embeddings
        embeddings = self.LayerNorm(embeddings)
        embeddings = self.dropout(embeddings)
        return embeddings


class CLIPAttention(nn.Module):
    """Multi-headed attention from 'Attention Is All You Need' paper"""

    def __init__(self, config):
        super().__init__()
        self.config = config
        self.embed_dim = config.hidden_size
        self.num_heads = config.num_attention_heads
        self.head_dim = self.embed_dim // self.num_heads
        assert (
            self.head_dim * self.num_heads == self.embed_dim
        ), f"embed_dim must be divisible by num_heads (got `embed_dim`: {self.embed_dim} and `num_heads`: {self.num_heads})."
        self.scale = self.head_dim ** -0.5
        self.dropout = config.attention_dropout

        self.k_proj = nn.Linear(self.embed_dim, self.embed_dim)
        self.v_proj = nn.Linear(self.embed_dim, self.embed_dim)
        self.q_proj = nn.Linear(self.embed_dim, self.embed_dim)
        self.out_proj = nn.Linear(self.embed_dim, self.embed_dim)

    def _shape(self, tensor: torch.Tensor, seq_len: int, bsz: int):
        return tensor.view(bsz, seq_len, self.num_heads, self.head_dim).transpose(1, 2).contiguous()

    def forward(
        self,
        hidden_states: torch.Tensor,
        output_attentions: bool = False,
        past_key_values: torch.Tensor = None,
        current_layer: int = None,
        output_qks=None,
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor], Optional[Tuple[torch.Tensor]]]:
        """Input shape: Batch x Time x Channel"""

        bsz, tgt_len, embed_dim = hidden_states.size()

        # get query proj
        query_states = self.q_proj(hidden_states) * self.scale
        key_states = self._shape(self.k_proj(hidden_states), -1, bsz)
        value_states = self._shape(self.v_proj(hidden_states), -1, bsz)
        qks = (key_states, value_states) if output_qks else None


        if past_key_values is not None:
            key_states = torch.cat([past_key_values[0], key_states], dim=2)
            value_states = torch.cat([past_key_values[1], value_states], dim=2)

        proj_shape = (bsz * self.num_heads, -1, self.head_dim)
        query_states = self._shape(query_states, tgt_len, bsz)

        query_states = query_states.view(*proj_shape)
        key_states = key_states.view(*proj_shape)
        value_states = value_states.view(*proj_shape)
        
        src_len = key_states.size(1)
        attn_weights = torch.bmm(query_states, key_states.transpose(1, 2))

        if attn_weights.size() != (bsz * self.num_heads, tgt_len, src_len):
            raise ValueError(
                f"Attention weights should be of size {(bsz * self.num_heads, tgt_len, src_len)}, but is {attn_weights.size()}"
            )       
        attn_weights = nn.functional.softmax(attn_weights, dim=-1)

        if output_attentions:
            # this operation is a bit akward, but it's required to
            # make sure that attn_weights keeps its gradient.
            # In order to do so, attn_weights have to reshaped
            # twice and have to be reused in the following
            attn_weights_reshaped = attn_weights.view(bsz, self.num_heads, tgt_len, src_len)
            attn_weights = attn_weights_reshaped.view(bsz * self.num_heads, tgt_len, src_len)
        else:
            attn_weights_reshaped = None

        attn_probs = nn.functional.dropout(attn_weights, p=self.dropout, training=self.training)

        attn_output = torch.bmm(attn_probs, value_states)

        if attn_output.size() != (bsz * self.num_heads, tgt_len, self.head_dim):
            raise ValueError(
                f"`attn_output` should be of size {(bsz, self.num_heads, tgt_len, self.head_dim)}, but is {attn_output.size()}"
            )

        attn_output = attn_output.view(bsz, self.num_heads, tgt_len, self.head_dim)
        attn_output = attn_output.transpose(1, 2)
        attn_output = attn_output.reshape(bsz, tgt_len, embed_dim)

        attn_output = self.out_proj(attn_output)

        return attn_output, attn_weights_reshaped, qks

def quick_gelu(x):
    return x * torch.sigmoid(1.702 * x)

class CLIPMLP(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.config = config
        self.activation_fn = quick_gelu
        self.fc1 = nn.Linear(config.hidden_size, config.intermediate_size)
        self.fc2 = nn.Linear(config.intermediate_size, config.hidden_size)

    def forward(self, hidden_states):
        hidden_states = self.fc1(hidden_states)
        hidden_states = self.activation_fn(hidden_states)
        hidden_states = self.fc2(hidden_states)
        return hidden_states


class BertSelfAttention(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.num_attention_heads = config.num_attention_heads   # 12
        self.attention_head_size = int(config.hidden_size / config.num_attention_heads) # 64
        self.all_head_size = self.num_attention_heads * self.attention_head_size    # 768

        self.query = nn.Linear(config.hidden_size, self.all_head_size)
        self.key = nn.Linear(config.hidden_size, self.all_head_size)
        self.value = nn.Linear(config.hidden_size, self.all_head_size)

        self.dropout = nn.Dropout(config.attention_probs_dropout_prob)

    def transpose_for_scores(self, x):
        new_x_shape = x.size()[:-1] + (self.num_attention_heads, self.attention_head_size)
        x = x.view(*new_x_shape)
        return x.permute(0, 2, 1, 3)

    def forward(
        self,
        hidden_states,
        attention_mask=None,
        head_mask=None,
        output_attentions=False,
        visual_hidden_state=None,
        output_qks=None,
        current_layer=None,
        past_key_values=None,
    ):
        mixed_query_layer = self.query(hidden_states)

        # If this is instantiated as a cross-attention module, the keys
        # and values come from an encoder; the attention mask needs to be
        # such that the encoder's padding tokens are not attended to.
        key_layer = self.transpose_for_scores(self.key(hidden_states))
        value_layer = self.transpose_for_scores(self.value(hidden_states))
        query_layer = self.transpose_for_scores(mixed_query_layer)

        qks = (key_layer, value_layer) if output_qks else None

        if past_key_values is not None:
            key_layer = torch.cat([past_key_values[0], key_layer], dim=2)
            value_layer = torch.cat([past_key_values[0], value_layer], dim=2)
        # Take the dot product between "query" and "key" to get the raw attention scores.
        attention_scores = torch.matmul(query_layer, key_layer.transpose(-1, -2))
        attention_scores = attention_scores / math.sqrt(self.attention_head_size)
        if attention_mask is not None:
            # Apply the attention mask is (precomputed for all layers in BertModel forward() function)
            bsz, nheads, length, dsize = past_key_values[0].size()
            visual_attention_mask = torch.ones((bsz, 1, 1, length)).to(attention_mask.device)  # bsz, 12, len, 64
            attention_mask = torch.cat((visual_attention_mask, attention_mask), dim=-1)
            attention_scores = attention_scores + attention_mask
        # Normalize the attention scores to probabilities.
        attention_probs = nn.Softmax(dim=-1)(attention_scores)

        # This is actually dropping out entire tokens to attend to, which might
        # seem a bit unusual, but is taken from the original Transformer paper.
        attention_probs = self.dropout(attention_probs)

        # Mask heads if we want to
        if head_mask is not None:
            attention_probs = attention_probs * head_mask
        context_layer = torch.matmul(attention_probs, value_layer)

        context_layer = context_layer.permute(0, 2, 1, 3).contiguous()
        new_context_layer_shape = context_layer.size()[:-2] + (self.all_head_size,)
        context_layer = context_layer.view(*new_context_layer_shape)    # bsz, 128, 768
        
        outputs = (context_layer, attention_probs) if output_attentions else (context_layer,)

        return outputs, qks


class BertSelfOutput(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.dense = nn.Linear(config.hidden_size, config.hidden_size)
        self.LayerNorm = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)

    def forward(self, hidden_states, input_tensor):
        hidden_states = self.dense(hidden_states)
        hidden_states = self.dropout(hidden_states)
        hidden_states = self.LayerNorm(hidden_states + input_tensor)
        return hidden_states

class BertAttention(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.self = BertSelfAttention(config)
        self.output = BertSelfOutput(config)
        self.pruned_heads = set()

    def forward(
        self,
        hidden_states,
        attention_mask=None,
        head_mask=None,
        output_attentions=False,
        visual_hidden_state=None,
        output_qks=None,
        current_layer=None,
        past_key_values=None,
    ):
        self_outputs, qks = self.self(
            hidden_states,
            attention_mask,
            head_mask,
            output_attentions,
            visual_hidden_state,
            output_qks,
            current_layer,
            past_key_values,
        )
        attention_output = self.output(self_outputs[0], hidden_states)
        outputs = (attention_output,) + self_outputs[1:]  # add attentions if we output them
        return outputs, qks


class BertIntermediate(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.dense = nn.Linear(config.hidden_size, config.intermediate_size)
        if isinstance(config.hidden_act, str):
            self.intermediate_act_fn = ACT2FN[config.hidden_act]
        else:
            self.intermediate_act_fn = config.hidden_act

    def forward(self, hidden_states):
        hidden_states = self.dense(hidden_states)
        hidden_states = self.intermediate_act_fn(hidden_states)
        return hidden_states


class BertOutput(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.dense = nn.Linear(config.intermediate_size, config.hidden_size)
        self.LayerNorm = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)

    def forward(self, hidden_states, input_tensor):
        hidden_states = self.dense(hidden_states)
        hidden_states = self.dropout(hidden_states)
        hidden_states = self.LayerNorm(hidden_states + input_tensor)
        return hidden_states


class CLIPEncoderLayer(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.embed_dim = config.hidden_size
        self.self_attn = CLIPAttention(config)
        self.layer_norm1 = nn.LayerNorm(self.embed_dim)
        self.mlp = CLIPMLP(config)
        self.layer_norm2 = nn.LayerNorm(self.embed_dim)

    def forward(
        self,
        hidden_states: torch.Tensor,
        output_attentions: bool = False,
        past_key_values: torch.Tensor = None,
        current_layer: int = None,
        output_qks = None
    ):
        """
        Args:
            hidden_states (:obj:`torch.FloatTensor`): input to the layer of shape :obj:`(seq_len, batch, embed_dim)`
            attention_mask (:obj:`torch.FloatTensor`): attention mask of size
                :obj:`(batch, 1, tgt_len, src_len)` where padding elements are indicated by very large negative values.
            layer_head_mask (:obj:`torch.FloatTensor`): mask for attention heads in a given layer of size
                :obj:`(config.encoder_attention_heads,)`.
            output_attentions (:obj:`bool`, `optional`):
                Whether or not to return the attentions tensors of all attention layers. See ``attentions`` under
                returned tensors for more detail.
        """
        residual = hidden_states

        hidden_states = self.layer_norm1(hidden_states)
        hidden_states, attn_weights, qks = self.self_attn(
            hidden_states=hidden_states,
            output_attentions=output_attentions,
            past_key_values=past_key_values,
            output_qks=output_qks,
            current_layer=current_layer,
        )
        hidden_states = residual + hidden_states

        residual = hidden_states
        hidden_states = self.layer_norm2(hidden_states)
        hidden_states = self.mlp(hidden_states)
        hidden_states = residual + hidden_states

        outputs = (hidden_states,)

        if output_attentions:
            outputs += (attn_weights,)
        
        if output_qks:
            outputs += (qks, )
    
        return outputs


class BertLayer(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.chunk_size_feed_forward = config.chunk_size_feed_forward
        self.seq_len_dim = 1
        self.attention = BertAttention(config)
        self.add_cross_attention = config.add_cross_attention
        self.intermediate = BertIntermediate(config)
        self.output = BertOutput(config)

    def forward(
        self,
        hidden_states,
        attention_mask=None,
        head_mask=None,
        output_attentions=False,
        visual_hidden_state=None,
        output_qks=None,
        current_layer=None,
        past_key_values=None,
    ):
        # decoder uni-directional self-attention cached key/values tuple is at positions 1,2
        # self_attn_past_key_value = past_key_value[:2] if past_key_value is not None else None
        self_attention_outputs, qks = self.attention(
            hidden_states,
            attention_mask,
            head_mask,
            output_attentions=output_attentions,
            visual_hidden_state=visual_hidden_state,
            output_qks=output_qks,
            current_layer=current_layer,
            past_key_values=past_key_values,
        )
        attention_output = self_attention_outputs[0]

        outputs = self_attention_outputs[1:]  # add self attentions if we output attention weights

        layer_output = apply_chunking_to_forward(
            self.feed_forward_chunk, self.chunk_size_feed_forward, self.seq_len_dim, attention_output
        )
        outputs = (layer_output,) + outputs
        if output_qks: 
            outputs += (qks,)

        return outputs

    def feed_forward_chunk(self, attention_output):
        intermediate_output = self.intermediate(attention_output)
        layer_output = self.output(intermediate_output, attention_output)
        return layer_output


class IFAEncoder(nn.Module):
    def __init__(self, vision_config, text_config):
        super().__init__()
        self.vision_config = vision_config
        self.text_config = text_config

        self.vision_layers = nn.ModuleList([CLIPEncoderLayer(vision_config) for _ in range(vision_config.num_hidden_layers)])
        self.text_layer = nn.ModuleList([BertLayer(text_config) for _ in range(text_config.num_hidden_layers)])
    
    def forward(
        self,
        vision_embeds=None,
        text_embeds=None,
        attention_mask=None,
        head_mask=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
    ):
        assert self.vision_config.num_hidden_layers == self.text_config.num_hidden_layers

        all_vision_hidden_states = () if output_hidden_states else None
        all_text_hidden_states = () if output_hidden_states else None
        all_vision_attentions = () if output_attentions else None
        all_text_attentions = () if output_attentions else None
        
        vision_hidden_states = vision_embeds
        text_hidden_states = text_embeds
        for idx in range(self.vision_config.num_hidden_layers):
            if output_hidden_states:
                all_vision_hidden_states = all_vision_hidden_states + (vision_hidden_states, )
                all_text_hidden_states = all_text_hidden_states + (text_hidden_states, )
            
            # vision
            # TODO: 9-12 layers past text as pkv to vision
            output_qks = True
            if idx == 0:
                bsz, length, dsize = text_embeds.size()
                visual_past_key_values = (text_embeds.view(bsz, 12, length, dsize//12),
                                          text_embeds.view(bsz, 12, length, dsize//12))
            else:
                visual_past_key_values = text_layer_output[-1]
            vision_layer_module = self.vision_layers[idx]
            vision_layer_output = vision_layer_module(
                    vision_hidden_states,
                    output_attentions=output_attentions,
                    past_key_values=visual_past_key_values,
                    current_layer=idx,
                    output_qks=output_qks,
            )
            vision_hidden_states = vision_layer_output[0]

            # text
            # TODO: 9-12 layers past vison qks to text
            if idx == 0:
                bsz, length, dsize = vision_embeds.size()
                text_past_key_values = (vision_embeds.view(bsz, 12, length, dsize//12),
                                          vision_embeds.view(bsz, 12, length, dsize//12))
            else:
                text_past_key_values = vision_layer_output[-1]

            layer_head_mask = head_mask[idx] if head_mask is not None else None
            text_layer_module = self.text_layer[idx]
            text_layer_output = text_layer_module(
                    text_hidden_states,
                    attention_mask=attention_mask,
                    head_mask=layer_head_mask,
                    visual_hidden_state=None,
                    past_key_values=text_past_key_values,
                    output_attentions=output_attentions,
                    output_qks=output_qks,
                    current_layer=idx,
            )
            text_hidden_states = text_layer_output[0]
            if output_attentions:
                all_vision_attentions = all_vision_attentions + (vision_layer_output[1], )
                all_text_attentions = all_text_attentions + (text_layer_output[1], )
        
        if output_hidden_states:
                all_vision_hidden_states = all_vision_hidden_states + (vision_hidden_states, )
                all_text_hidden_states = all_text_hidden_states + (text_hidden_states, )
        
        if not return_dict:
            return tuple(
                v for v in [
                    text_hidden_states,
                    all_text_hidden_states,
                    all_text_attentions,
                ] if v is not None)
        return BaseModelOutput(
            last_hidden_state=text_hidden_states, hidden_states=all_text_hidden_states, attentions=all_text_attentions
        )


class BertPooler(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.dense = nn.Linear(config.hidden_size, config.hidden_size)
        self.activation = nn.Tanh()

    def forward(self, hidden_states):
        # We "pool" the model by simply taking the hidden state corresponding
        # to the first token.
        first_token_tensor = hidden_states[:, 0]
        pooled_output = self.dense(first_token_tensor)
        pooled_output = self.activation(pooled_output)
        return pooled_output


class IFAModel(nn.Module):
    def __init__(self, vision_config, text_config, add_pooling_layer=True):
        super(IFAModel, self).__init__()
        # vision model
        self.vision_config = vision_config
        self.vision_embeddings = CLIPVisionEmbeddings(vision_config)
        self.vision_pre_layrnorm = nn.LayerNorm(vision_config.hidden_size)
        self.vision_post_layernorm = nn.LayerNorm(vision_config.hidden_size)

        # text model
        self.text_config = text_config
        self.text_embeddings = BertEmbeddings(text_config)
        self.text_pooler = BertPooler(text_config) if add_pooling_layer else None

        # all
        self.encoder = IFAEncoder(vision_config, text_config)

        self.device = vision_config.device

    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        token_type_ids=None,
        position_ids=None,
        head_mask=None,
        
        pixel_values=None,
        aux_values=None, 
        rcnn_values=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
    ):
        # pre vision
        vision_embedding_output = self.vision_embeddings(pixel_values, aux_values, rcnn_values)
        vision_embedding_output = self.vision_pre_layrnorm(vision_embedding_output)

        # pre text
        input_shape = input_ids.size()
        batch_size, seq_length = input_shape
        device = input_ids.device
        if attention_mask is None:
            attention_mask = torch.ones(((batch_size, seq_length)), device=device)
        if token_type_ids is None:
            raise ValueError("token_type_ids is None!")

        extended_attention_mask: torch.Tensor = get_extended_attention_mask(attention_mask, input_shape, device)
        head_mask = get_head_mask(head_mask, self.text_config.num_hidden_layers)    # [None]*12

        text_embedding_output = self.text_embeddings(
            input_ids=input_ids,
            position_ids=position_ids,
            token_type_ids=token_type_ids,
        )

        # all encoder
        encoder_outputs = self.encoder(
            vision_embeds=vision_embedding_output,
            text_embeds=text_embedding_output,
            attention_mask=extended_attention_mask,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )
        sequence_output = encoder_outputs[0]
        pooled_output = self.text_pooler(sequence_output) if self.text_pooler is not None else None

        if not return_dict:
            return (sequence_output, pooled_output) + encoder_outputs[1:]

        return BaseModelOutputWithPooling(
            last_hidden_state=sequence_output,
            pooler_output=pooled_output,
            hidden_states=encoder_outputs.hidden_states,
            attentions=encoder_outputs.attentions,
        )

    def _init_text_weights(self, module):
        """Initialize the weights"""
        if isinstance(module, nn.Linear):
            # Slightly different from the TF version which uses truncated_normal for initialization
            # cf https://github.com/pytorch/pytorch/pull/5617
            module.weight.data.normal_(mean=0.0, std=self.text_config.initializer_range)
            if module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.Embedding):
            module.weight.data.normal_(mean=0.0, std=self.text_config.initializer_range)
            if module.padding_idx is not None:
                module.weight.data[module.padding_idx].zero_()
        elif isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)

    def get_input_embeddings(self):
        return self.text_embeddings.word_embeddings

    def set_input_embeddings(self, value):
        self.text_embeddings.word_embeddings = value

    def resize_token_embeddings(self, new_num_tokens):
        old_embeddings = self.get_input_embeddings()
        new_embeddings = self._get_resized_embeddings(old_embeddings, new_num_tokens)
        self.set_input_embeddings(new_embeddings)

    def _get_resized_embeddings(
        self, old_embeddings: nn.Embedding, new_num_tokens: Optional[int] = None
    ) -> nn.Embedding:
        """
        Build a resized Embedding Module from a provided token Embedding Module. Increasing the size will add newly
        initialized vectors at the end. Reducing the size will remove vectors from the end

        Args:
            old_embeddings (:obj:`torch.nn.Embedding`):
                Old embeddings to be resized.
            new_num_tokens (:obj:`int`, `optional`):
                New number of tokens in the embedding matrix.

                Increasing the size will add newly initialized vectors at the end. Reducing the size will remove
                vectors from the end. If not provided or :obj:`None`, just returns a pointer to the input tokens
                :obj:`torch.nn.Embedding`` module of the model without doing anything.

        Return:
            :obj:`torch.nn.Embedding`: Pointer to the resized Embedding Module or the old Embedding Module if
            :obj:`new_num_tokens` is :obj:`None`
        """
        if new_num_tokens is None:
            return old_embeddings
        else:
            old_num_tokens, old_embedding_dim = old_embeddings.weight.size()

        if old_num_tokens == new_num_tokens:
            return old_embeddings

        if not isinstance(old_embeddings, nn.Embedding):
            raise TypeError(
                f"Old embeddings are of type {type(old_embeddings)}, which is not an instance of {nn.Embedding}."
                f"You should either use a different resize function or make sure that `old_embeddings` are an instance of {nn.Embedding}."
            )

        # Build new embeddings
        new_embeddings = nn.Embedding(new_num_tokens, old_embedding_dim).to(
            self.device, dtype=old_embeddings.weight.dtype
        )

        # initialize all new embeddings (in particular added tokens)
        self._init_text_weights(new_embeddings)

        # Copy token embeddings from the previous weights

        # numbers of tokens to copy
        n = min(old_num_tokens, new_num_tokens)
        new_embeddings.weight.data[:n, :] = old_embeddings.weight.data[:n, :]

        return new_embeddings


import torch
import random
import logging
import numpy as np
from typing import List, Tuple, Dict, Union

logger = logging.getLogger(__name__)

__all__ = [
    'manual_seed',
    'seq_len_to_mask',
    'to_one_hot',
]


def manual_seed(seed: int = 1) -> None:
    """
        设置seed。
    """
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    #if torch.cuda.CUDA_ENABLED and use_deterministic_cudnn:
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False



def seq_len_to_mask(seq_len: Union[List, np.ndarray, torch.Tensor], max_len=512, mask_pos_to_true=True):
    """
    将一个表示sequence length的一维数组转换为二维的mask，默认pad的位置为1。
    转变 1-d seq_len到2-d mask。

    Args :
        seq_len (list, np.ndarray, torch.LongTensor) : shape将是(B,)
        max_len (int): 将长度pad到这个长度。默认(None)使用的是seq_len中最长的长度。但在nn.DataParallel的场景下可能不同卡的seq_len会有区别，所以需要传入一个max_len使得mask的长度是pad到该长度。
    Return: 
        mask (np.ndarray, torch.Tensor) : shape将是(B, max_length)， 元素类似为bool或torch.uint8
    """
    if isinstance(seq_len, list):
        seq_len = np.array(seq_len)

    if isinstance(seq_len, np.ndarray):
        seq_len = torch.from_numpy(seq_len)

    if isinstance(seq_len, torch.Tensor):
        assert seq_len.dim() == 1, logger.error(f"seq_len can only have one dimension, got {seq_len.dim()} != 1.")
        batch_size = seq_len.size(0)
        max_len = int(max_len) if max_len else seq_len.max().long()
        broad_cast_seq_len = torch.arange(max_len).expand(batch_size, -1).to(seq_len.device)
        if mask_pos_to_true:
            mask = broad_cast_seq_len.ge(seq_len.unsqueeze(1))
        else:
            mask = broad_cast_seq_len.lt(seq_len.unsqueeze(1))
    else:
        raise logger.error("Only support 1-d list or 1-d numpy.ndarray or 1-d torch.Tensor.")

    return mask


def to_one_hot(x: torch.Tensor, length: int) -> torch.Tensor:
    """
    Args:
        x (torch.Tensor):[B] , 一般是 target 的值
        length (int) : L ,一般是关系种类树
    Return:
        x_one_hot.to(device=x.device) (torch.Tensor) : [B, L]  每一行，只有对应位置为1，其余为0
    """
    B = x.size(0)
    x_one_hot = torch.zeros(B, length)
    for i in range(B):
        x_one_hot[i, x[i]] = 1.0

    return x_one_hot.to(device=x.device)


import os
from copy import deepcopy
import torch
from torch import nn
from .modeling_t5 import T5ForConditionalGeneration

class T5Prompt(nn.Module):
    def __init__(self, model_name_or_path, config, args):
        super().__init__()
        self.t5 = T5ForConditionalGeneration.from_pretrained(model_name_or_path, config=config)
        self.config = self.t5.config
        self.args = args
        
        # model config related
        self.match_n_layer = self.config.num_decoder_layers
        self.match_n_head = self.config.num_heads
        self.n_embd = self.config.d_model
        self.match_n_embd = self.config.d_kv
        
        # prefix related
        self.prompt_len = args.prompt_len
        self.prompt_dim = args.prompt_dim
        self.prompt_inputs = torch.arange(self.prompt_len).long()    # [0, 1, ..., prompt_len-1]
        
        self.wte = nn.Embedding(self.prompt_len, self.n_embd)
        self.control_trans = nn.Sequential(
            nn.Linear(self.n_embd, self.prompt_dim),
            nn.Tanh(),
            nn.Linear(self.prompt_dim, self.match_n_layer * 2 * self.match_n_head * self.match_n_embd),
        )
        
        self.wte_enc = nn.Embedding(self.prompt_len, self.n_embd)
        self.control_trans_enc = nn.Sequential(
            nn.Linear(self.n_embd, self.prompt_dim),
            nn.Tanh(),
            nn.Linear(self.prompt_dim, self.match_n_layer * 2 * self.match_n_head * self.match_n_embd),
        )
        
        self.wte_dec = nn.Embedding(self.prompt_len, self.n_embd)
        self.control_trans_dec = nn.Sequential(
            nn.Linear(self.n_embd, self.prompt_dim),
            nn.Tanh(),
            nn.Linear(self.prompt_dim, self.match_n_layer * 2 * self.match_n_head * self.match_n_embd),
        )
        
        self.dropout = nn.Dropout(0.1)
                    
        if args.source_prefix_path is not None and 'concat' in args.prefix_fusion_way:
            self.source_prompt_inputs = torch.arange(self.prompt_len).long()    # [0, 1, ..., prompt_len-1]
            self.source_wte = nn.Embedding(self.prompt_len, self.n_embd)
            self.source_control_trans = nn.Sequential(
                nn.Linear(self.n_embd, self.prompt_dim),
                nn.Tanh(),
                nn.Linear(self.prompt_dim, self.match_n_layer * 2 * self.match_n_head * self.match_n_embd),
            )
            
            self.source_wte_enc = nn.Embedding(self.prompt_len, self.n_embd)
            self.source_control_trans_enc = nn.Sequential(
                nn.Linear(self.n_embd, self.prompt_dim),
                nn.Tanh(),
                nn.Linear(self.prompt_dim, self.match_n_layer * 2 * self.match_n_head * self.match_n_embd),
            )
            
            self.source_wte_dec = nn.Embedding(self.prompt_len, self.n_embd)
            self.source_control_trans_dec = nn.Sequential(
                nn.Linear(self.n_embd, self.prompt_dim),
                nn.Tanh(),
                nn.Linear(self.prompt_dim, self.match_n_layer * 2 * self.match_n_head * self.match_n_embd),
            )
            
        if args.multi_source_path:
            # module
            self.prefix_project_layer = nn.Sequential(
                nn.Linear(self.n_embd, 500),
                nn.Tanh(),
                nn.Linear(500, self.n_embd)
            )
                    
        if 'lstm' in args.prefix_fusion_way:
            self.lstm = nn.LSTM(input_size=self.match_n_head * self.match_n_embd, 
                                hidden_size=self.match_n_head * self.match_n_embd, 
                                num_layers=2, 
                                batch_first=True,
                                bidirectional=True)
            
        if args.freeze_plm:
            for name, param in self.t5.named_parameters():
                if 'encoder' in name or 'decoder' in name and 'adapter' not in name:
                    param.requires_grad = False
                    
        if args.freeze_prefix:
            for name, param in self.named_parameters():
                if 'wte' in name or 'control_trans' in name:
                    param.requires_grad = False
       
    def get_prompt(self, bsz=None):
        input_tokens = self.prompt_inputs.unsqueeze(0).expand(bsz, -1).to(self.t5.device)
        temp_control = self.wte(input_tokens)
        past_key_values = self.control_trans(temp_control)  # bsz, seqlen, layer*emb

        bsz, seqlen, _ = past_key_values.shape
        past_key_values = past_key_values.view(
            bsz, seqlen, self.match_n_layer * 2, self.match_n_head, self.match_n_embd
        )
        past_key_values = self.dropout(past_key_values)
        past_key_values = past_key_values.permute([2, 0, 3, 1, 4]).split(2)

        # Cross prefix
        temp_control_dec = self.wte_dec(input_tokens)
        past_key_values_dec = self.control_trans_dec(
            temp_control_dec
        )  # bsz, seqlen, layer*emb

        bsz, seqlen, _ = past_key_values_dec.shape
        past_key_values_dec = past_key_values_dec.view(
            bsz, seqlen, self.match_n_layer * 2, self.match_n_head, self.match_n_embd
        )
        past_key_values_dec = self.dropout(past_key_values_dec)
        past_key_values_dec = past_key_values_dec.permute([2, 0, 3, 1, 4]).split(2)

        # Encoder prefix
        input_tokens_enc = (self.prompt_inputs.unsqueeze(0).expand(bsz, -1)).to(self.t5.device)
        temp_control_enc = self.wte_enc(input_tokens_enc)
        past_key_values_enc = self.control_trans_enc(
            temp_control_enc
        )  # bsz, seqlen, layer*emb

        bsz_enc, seqlen, _ = past_key_values_enc.shape
        past_key_values_enc = past_key_values_enc.view(
            bsz_enc,
            seqlen,
            self.match_n_layer * 2,
            self.match_n_head,
            self.match_n_embd,
        )
        past_key_values_enc = self.dropout(past_key_values_enc)
        past_key_values_enc = past_key_values_enc.permute([2, 0, 3, 1, 4]).split(2)

        result = []
        for i, key_val in enumerate(past_key_values):
            temp = dict()
            temp["decoder_prompt"] = {
                "prev_key": key_val[0].contiguous(),
                "prev_value": key_val[1].contiguous(),
                "prev_key_padding_mask": torch.zeros(bsz, seqlen)
                    .to(key_val.device)
                    .bool()
                # bsz, prompt_len
            }
            key_val_dec = past_key_values_dec[i]
            temp["cross_attention_prompt"] = {
                "prev_key": key_val_dec[0].contiguous(),
                "prev_value": key_val_dec[1].contiguous(),
                "prev_key_padding_mask": torch.zeros(bsz, seqlen)
                    .to(key_val_dec.device)
                    .bool(),
            }
            key_val_enc = past_key_values_enc[i]
            temp["encoder_prompt"] = {
                "prev_key": key_val_enc[0].contiguous(),
                "prev_value": key_val_enc[1].contiguous(),
                "prev_key_padding_mask": torch.zeros(bsz_enc, seqlen)
                    .to(key_val_enc.device)
                    .bool(),
            }
            result.append(temp)

        return result
    
    
    def get_source_prompt(self, bsz=None):
        input_tokens = self.source_prompt_inputs.unsqueeze(0).expand(bsz, -1).to(self.t5.device)
        temp_control = self.source_wte(input_tokens)
        past_key_values = self.source_control_trans(temp_control)  # bsz, seqlen, layer*emb

        bsz, seqlen, _ = past_key_values.shape
        past_key_values = past_key_values.view(
            bsz, seqlen, self.match_n_layer * 2, self.match_n_head, self.match_n_embd
        )
        past_key_values = self.dropout(past_key_values)
        past_key_values = past_key_values.permute([2, 0, 3, 1, 4]).split(2)

        # Cross prefix
        temp_control_dec = self.source_wte_dec(input_tokens)
        past_key_values_dec = self.source_control_trans_dec(
            temp_control_dec
        )  # bsz, seqlen, layer*emb

        bsz, seqlen, _ = past_key_values_dec.shape
        past_key_values_dec = past_key_values_dec.view(
            bsz, seqlen, self.match_n_layer * 2, self.match_n_head, self.match_n_embd
        )
        past_key_values_dec = self.dropout(past_key_values_dec)
        past_key_values_dec = past_key_values_dec.permute([2, 0, 3, 1, 4]).split(2)

        # Encoder prefix
        input_tokens_enc = (self.source_prompt_inputs.unsqueeze(0).expand(bsz, -1)).to(self.t5.device)
        temp_control_enc = self.source_wte_enc(input_tokens_enc)
        past_key_values_enc = self.source_control_trans_enc(
            temp_control_enc
        )  # bsz, seqlen, layer*emb

        bsz_enc, seqlen, _ = past_key_values_enc.shape
        past_key_values_enc = past_key_values_enc.view(
            bsz_enc,
            seqlen,
            self.match_n_layer * 2,
            self.match_n_head,
            self.match_n_embd,
        )
        past_key_values_enc = self.dropout(past_key_values_enc)
        past_key_values_enc = past_key_values_enc.permute([2, 0, 3, 1, 4]).split(2)

        result = []
        for i, key_val in enumerate(past_key_values):
            temp = dict()
            temp["decoder_prompt"] = {
                "prev_key": key_val[0].contiguous(),
                "prev_value": key_val[1].contiguous(),
                "prev_key_padding_mask": torch.zeros(bsz, seqlen)
                    .to(key_val.device)
                    .bool()
                # bsz, prompt_len
            }
            key_val_dec = past_key_values_dec[i]
            temp["cross_attention_prompt"] = {
                "prev_key": key_val_dec[0].contiguous(),
                "prev_value": key_val_dec[1].contiguous(),
                "prev_key_padding_mask": torch.zeros(bsz, seqlen)
                    .to(key_val_dec.device)
                    .bool(),
            }
            key_val_enc = past_key_values_enc[i]
            temp["encoder_prompt"] = {
                "prev_key": key_val_enc[0].contiguous(),
                "prev_value": key_val_enc[1].contiguous(),
                "prev_key_padding_mask": torch.zeros(bsz_enc, seqlen)
                    .to(key_val_enc.device)
                    .bool(),
            }
            result.append(temp)

        return result
    
    def convert_prefix(self, prefix_dict):
        '''trans dict prefix to 1-D tensor prefix'''
        trans_prefix = []
        for i, prefix in enumerate(prefix_dict): # 12 layer
            # 'decoder_prompt', 'cross_attention_prompt', 'encoder_prompt'
            current_key_prefix = []
            for key in prefix.keys():
                # prev_key', 'prev_value', 'prev_key_padding_mask'
                current_key_prefix.append(torch.stack([
                            prefix[key]['prev_key'].squeeze(0).transpose(1, 0).reshape(-1, self.n_embd).mean(0),
                            prefix[key]['prev_value'].squeeze(0).transpose(1, 0).reshape(-1, self.n_embd).mean(0),
                        ], dim=0).mean(0))  #  768
            trans_prefix.append(torch.stack(current_key_prefix, dim=0).mean(0))
        trans_prefix = torch.stack(trans_prefix, dim=0).mean(0)
        return trans_prefix


    def normalize_multi_source(self, label_word, tokenizer, device):
        '''This function is to normalize multi source to aggregate them.'''
        assert self.args.multi_source_path is not None
        
        # load multi source prefix and label word
        source_paths = self.args.multi_source_path.split(',')
        self.source_prefixes = [torch.load(os.path.join(path, 'prefix.pt')) for path in source_paths]
                
        trans_source_prefixes = []  # n x 768
        for source_prefix in self.source_prefixes: # n source
            trans_source_prefixes.append(self.convert_prefix(source_prefix))
        self.trans_source_prefixes = torch.stack(trans_source_prefixes, dim=0)
        
        self.trans_target_prefix = self.convert_prefix(self.get_prompt(bsz=1)).unsqueeze(0)
            
        source_label_words = [torch.load(os.path.join(path, 'label_word.pt')) for path in source_paths]
        for i in range(len(source_label_words)):
            source_label_words[i] = torch.stack(list(source_label_words[i].values()), dim=0).mean(0)
        self.source_label_words = torch.stack(source_label_words, dim=0).to(device)
        self.label_word_id = [tokenizer.encode(label, add_special_tokens=False) for label in label_word]
    
    
    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        labels=None,
        decoder_input_ids=None,
        return_dict=True,
    ):
        bsz = input_ids.shape[0]
        past_prompt = self.get_prompt(bsz=bsz)
        
        if self.args.multi_source_path:
            # 1. compute cosine similarity between target prefix and source prefixes
            trans_source_prefixes = self.prefix_project_layer(self.trans_source_prefixes.detach().to(input_ids.device))   # project
            
            trans_target_prefix = self.convert_prefix(self.get_prompt(bsz=1))
            trans_target_prefix = self.prefix_project_layer(trans_target_prefix).detach().unsqueeze(0)
            
            prefix_sim = torch.cosine_similarity(trans_target_prefix.to(input_ids.device), trans_source_prefixes.to(input_ids.device))   # n
            prefix_sim = prefix_sim / prefix_sim.sum()
            
            # 2. compute cosine similarity between target label word and source label words
            target_label_word = []
            for label_id in self.label_word_id:
                if len(label_id) > 1:
                    target_label_word.append(torch.stack([self.t5.shared.weight.data[id] for id in label_id], dim=0).mean(0))
                else:
                    target_label_word.append(self.t5.shared.weight.data[label_id[0]])
            target_label_word = torch.stack(target_label_word, dim=0).mean(0).to(input_ids.device)
            label_word_sim = torch.cosine_similarity(target_label_word.unsqueeze(0), self.source_label_words.to(input_ids.device))
            label_word_sim = label_word_sim / label_word_sim.sum()
            
            total_sim = (prefix_sim + label_word_sim) / 2

            # aggregate
            for i, prefix in enumerate(past_prompt):    # 24
                for key in prefix.keys():               # 'decoder_prompt', 'cross_attention_prompt', 'encoder_prompt'
                    for sub_key in prefix[key].keys():  # prev_key', 'prev_value', 'prev_key_padding_mask'
                        if len(prefix[key][sub_key].size()) == 4:
                            agg_prefix = torch.zeros((1, self.match_n_head, self.prompt_len, self.match_n_embd), device=input_ids.device)
                            for j in range(len(self.source_prefixes)):
                                agg_prefix += self.source_prefixes[j][i][key][sub_key].to(input_ids.device) * total_sim[j]
                            past_prompt[i][key][sub_key] = (past_prompt[i][key][sub_key] + agg_prefix) / 2

        if 'concat' in self.args.prefix_fusion_way:
            source_past_prompt = self.get_source_prompt(bsz=bsz)
            for i, prefix in enumerate(past_prompt):    # 24
                for key in prefix.keys():               # 'decoder_prompt', 'cross_attention_prompt', 'encoder_prompt'
                    for sub_key in prefix[key].keys():  # prev_key', 'prev_value', 'prev_key_padding_mask'
                        if len(prefix[key][sub_key].size()) == 4:
                            past_prompt[i][key][sub_key] = torch.cat((past_prompt[i][key][sub_key], 
                                                                      source_past_prompt[i][key][sub_key]), dim=2)
                        elif len(prefix[key][sub_key].size()) == 2:
                            past_prompt[i][key][sub_key] = torch.cat((past_prompt[i][key][sub_key], 
                                                                      source_past_prompt[i][key][sub_key]), dim=1)
        
        if 'lstm' in self.args.prefix_fusion_way:
            for i, prefix in enumerate(past_prompt):    # 24
                for key in prefix.keys():               # 'decoder_prompt', 'cross_attention_prompt', 'encoder_prompt'
                    for sub_key in prefix[key].keys():  # prev_key', 'prev_value', 'prev_key_padding_mask'
                        if len(prefix[key][sub_key].size()) == 4:
                            _, _, prefix_len, _ = prefix[key][sub_key].size()
                            lstm_output = self.lstm(past_prompt[i][key][sub_key].reshape(bsz, prefix_len, self.match_n_head * self.match_n_embd))[0]
                            if lstm_output.size(-1) == 2 * (self.match_n_head * self.match_n_embd):   # bidirectional
                                past_prompt[i][key][sub_key] = lstm_output.reshape(bsz, self.match_n_head, prefix_len, 2, self.match_n_embd).mean(dim=-2)
                            else:
                                past_prompt[i][key][sub_key] = lstm_output.reshape(bsz, self.match_n_head, prefix_len, self.match_n_embd)
      

        return self.t5(
            input_ids=input_ids,
            attention_mask=attention_mask,
            labels=labels,
            decoder_input_ids=decoder_input_ids,
            return_dict=return_dict,
            past_prompt=past_prompt,
        )
    
    
    def generate(
        self,
        input_ids,
        attention_mask,
        labels,
        **kwargs,
    ):
        bsz = input_ids.shape[0]
        past_prompt = self.get_prompt(bsz=bsz)
        
        if self.args.multi_source_path:
            # 1. compute cosine similarity between target prefix and source prefixes
            trans_source_prefixes = self.prefix_project_layer(self.trans_source_prefixes.detach().to(input_ids.device))   # project
            
            trans_target_prefix = self.convert_prefix(self.get_prompt(bsz=1))
            trans_target_prefix = self.prefix_project_layer(trans_target_prefix).detach().unsqueeze(0)
            
            prefix_sim = torch.cosine_similarity(trans_target_prefix.to(input_ids.device), trans_source_prefixes.to(input_ids.device))   # n
            prefix_sim = prefix_sim / prefix_sim.sum()
            
            # 2. compute cosine similarity between target label word and source label words
            target_label_word = []
            for label_id in self.label_word_id:
                if len(label_id) > 1:
                    target_label_word.append(torch.stack([self.t5.shared.weight.data[id] for id in label_id], dim=0).mean(0))
                else:
                    target_label_word.append(self.t5.shared.weight.data[label_id[0]])
            target_label_word = torch.stack(target_label_word, dim=0).mean(0).to(input_ids.device)
            label_word_sim = torch.cosine_similarity(target_label_word.unsqueeze(0), self.source_label_words.to(input_ids.device))
            label_word_sim = label_word_sim / label_word_sim.sum()
            
            total_sim = (prefix_sim + label_word_sim) / 2

            # aggregate
            for i, prefix in enumerate(past_prompt):    # 24
                for key in prefix.keys():               # 'decoder_prompt', 'cross_attention_prompt', 'encoder_prompt'
                    for sub_key in prefix[key].keys():  # prev_key', 'prev_value', 'prev_key_padding_mask'
                        if len(prefix[key][sub_key].size()) == 4:
                            agg_prefix = torch.zeros((1, self.match_n_head, self.prompt_len, self.match_n_embd), device=input_ids.device)
                            for j in range(len(self.source_prefixes)):
                                agg_prefix += self.source_prefixes[j][i][key][sub_key].to(input_ids.device) * total_sim[j]
                            past_prompt[i][key][sub_key] = (past_prompt[i][key][sub_key] + agg_prefix) / 2
          
          
        if 'concat' in self.args.prefix_fusion_way:
            source_past_prompt = self.get_source_prompt(bsz=bsz)
            for i, prefix in enumerate(past_prompt):    # 24
                for key in prefix.keys():               # 'decoder_prompt', 'cross_attention_prompt', 'encoder_prompt'
                    for sub_key in prefix[key].keys():  # prev_key', 'prev_value', 'prev_key_padding_mask'
                        if len(prefix[key][sub_key].size()) == 4:
                            past_prompt[i][key][sub_key] = torch.cat((past_prompt[i][key][sub_key], 
                                                                      source_past_prompt[i][key][sub_key]), dim=2)
                        elif len(prefix[key][sub_key].size()) == 2:
                            past_prompt[i][key][sub_key] = torch.cat((past_prompt[i][key][sub_key], 
                                                                      source_past_prompt[i][key][sub_key]), dim=1)
        
        if 'lstm' in self.args.prefix_fusion_way:
            for i, prefix in enumerate(past_prompt):    # 24
                for key in prefix.keys():               # 'decoder_prompt', 'cross_attention_prompt', 'encoder_prompt'
                    for sub_key in prefix[key].keys():  # prev_key', 'prev_value', 'prev_key_padding_mask'
                        if len(prefix[key][sub_key].size()) == 4:
                            _, _, prefix_len, _ = prefix[key][sub_key].size()
                            lstm_output = self.lstm(past_prompt[i][key][sub_key].reshape(bsz, prefix_len, self.match_n_head * self.match_n_embd))[0]
                            if lstm_output.size(-1) == 2 * (self.match_n_head * self.match_n_embd):   # bidirectional
                                past_prompt[i][key][sub_key] = lstm_output.reshape(bsz, self.match_n_head, prefix_len, 2, self.match_n_embd).mean(dim=-2)
                            else:
                                past_prompt[i][key][sub_key] = lstm_output.reshape(bsz, self.match_n_head, prefix_len, self.match_n_embd)
      
        
        generated_ids = self.t5.generate(
            input_ids=input_ids,
            attention_mask=attention_mask,
            past_prompt=past_prompt,
            use_cache=True,
            **kwargs,
        )

        return generated_ids